/* pokersol.c */
/* A simple Poker Solitaire game written in C for cc65 */
/* Bill Kendrick <bill@newbreedsoftware.com> */
/* June 19, 2015 - July 2, 2015 */

#define VERSION "1.1"

#include <atari.h>

#include <peekpoke.h>

#include "extended_headers/gtia.h"
#define X_GTIA_READ  (*(struct __x_gtia_read*)0xD000)
#define X_GTIA_WRITE (*(struct __x_gtia_write*)0xD000)

#include "extended_headers/pbi.h"

#include "extended_headers/pia.h"
#define X_PIA (*(struct __x_pia*)0xD300)

#include "extended_headers/antic.h"
#define X_ANTIC (*(struct __x_antic*)0xD400)

#include "extended_headers/pokey.h"
#define X_POKEY_READ  (*(struct __x_pokey_read*)0xD200)
#define X_POKEY_WRITE (*(struct __x_pokey_write*)0xD200)

#include "extended_headers/page0.h"
#include "extended_headers/page2.h"

#include "extended_headers/atascii.h"

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#include "lib/sound.h"

/*
Royal flush      Ah  Kh  Qh  Jh 10h  [straight flush, 10->Ace]
Straight flush   10h 9h  8h  7h  6h  [consecutive values, all same suit]
Four of a kind   Ah  Ad  As  Ac  3c  [four of same value]
Full house       Kd  Kc  8h  8c  8d  [2 of one value, 3 of another value]
Straight         9h  8c  7d  6d  5h  [consecutive values, not all same suit]
Flush            Kc  Qc  9c  8c  5c  [all same suit]
Three of a kind  Jd  Jh  Js  3c  8h  [three of same value]
Two pairs        Qd  Qh  7c  7h  3h  [2 of one value, 2 of another]
One pair         Ac  Ad  3c  8c 10h  [2 of same value]
*/

enum {
  HAND_NA,
  HAND_NOTHING,
  HAND_ONE_PAIR,
  HAND_TWO_PAIRS,
  HAND_THREE_KIND,
  HAND_FLUSH,
  HAND_STRAIGHT,
  HAND_FULL_HOUSE,
  HAND_FOUR_KIND,
  HAND_STRAIGHT_FLUSH,
  HAND_ROYAL_FLUSH,
  NUM_HANDS
};

/* Need to be small enough to fit on right of grid */
char * hand_names[NUM_HANDS] = {
  "",
  "NOTHING",
  "ONE PAIR",
  "TWO PAIRS",
  "3 OF A KIND",
  "STRAIGHT",
  "FLUSH",
  "FULL HOUSE",
  "4 OF A KIND",
  "STR. FLUSH",
  "ROYAL FLUSH",
};

/* Using American point system, per https://en.wikipedia.org/wiki/Poker_Squares */
unsigned int hand_scores[NUM_HANDS] = { 0, 0, 2, 5, 10, 15, 20, 25, 50, 75, 100 };

#define NUM_CARD_VALUES 13
#define NUM_CARD_SUITS  4
unsigned char card_values[NUM_CARD_VALUES + 1] = "A23456789*JQK";
unsigned char card_suits[NUM_CARD_SUITS + 1]   = CHS_SPADE CHS_HEART CHS_CLUB CHS_DIAMOND;

unsigned char deck[25];
unsigned char grid[25];

unsigned char * SC;

/* Shuffle a fresh deck of 52 cards (we only use 25) */
void shuffle_deck(void) {
  unsigned char sorted_deck[52];
  unsigned char i, c;

  for (i = 0; i < 52; i++) {
    sorted_deck[i] = i + 1;
  }

  for (i = 0; i < 25; i++) {
    c = X_POKEY_READ.random % (52 - i);
    deck[i] = sorted_deck[c];
    sorted_deck[c] = sorted_deck[51 - i];
  }
}

/* h/t http://www.atarimagazines.com/compute/issue70/113_1_INSIGHT_Atari.php */
void putch(unsigned char x, unsigned char y, unsigned char ch) {
  unsigned char inv;
  unsigned int pos;

  inv = (ch & 0x80);
  ch = (ch & 0x7F);
  
  if (ch < 32) {
    ch = inv + ch + 64;
  } else if (ch >= 96) {
    ch = inv + ch;
  } else {
    ch = inv + ch - 32;
  }

  pos = ((int) y) * 40 + x;

  SC[pos] = ch;
}


void print_card(unsigned char x, unsigned char y, unsigned char c) {
  if (c == 0) {
    putch(x, y, '-');
    putch(x + 1, y, '-');
    return;
  }

  c--;
  
  putch(x, y, card_values[c % 13] + 128);
  putch(x + 1, y, card_suits[c / 13] + 128);
}

typedef struct kind_type_t {
  unsigned char val;
  unsigned char cnt;
} kind_type;

unsigned char calc_hand(unsigned char * cards) {
  unsigned char i, j, sorted, tmp, not_full;
  unsigned char sorted_cards[5];
  unsigned char straight, flush;
  unsigned char num_kinds, cnt, found;
  kind_type kinds[5];

  /* Copy hand for sorting, and see if it's full */
  not_full = 0;
  for (i = 0; i < 5; i++) {
    sorted_cards[i] = cards[i] - 1;
    if (cards[i] == 0) {
      not_full = 1;
    }
  }

  /* Not a full hand?  No score yet */
  if (not_full) {
    return HAND_NA;
  }

  /* Sort the hand */
  do {
    sorted = 1;

    for (i = 0; i < 4; i++) {
      if ((sorted_cards[i] % 13) > (sorted_cards[i + 1] % 13)) {
        tmp = sorted_cards[i];
        sorted_cards[i] = sorted_cards[i + 1];
        sorted_cards[i + 1] = tmp;
        sorted = 0;
      }
    }
  } while (!sorted);

  /* Check for straight & flush */
  straight = 1;
  flush = 1;
  for (i = 1; i < 5; i++) {
    if ((sorted_cards[i] % 13) != (sorted_cards[i - 1] % 13) + 1) {
      if (i == 1 && (sorted_cards[i] % 13) == 9 && (sorted_cards[i - 1] & 13) == 0) {
        /* special case; A, 10, J, Q, K */
      } else {
        /* card[i] is not card[i-1]+1? not a straight */
        straight = 0;
      }
    }
    if (sorted_cards[i] / 13 != sorted_cards[i - 1] / 13) {
      /* card[i]'s suit is not card[i-1]'s suit? not a flush */
      flush = 0;
    }
  }

  /* Straights */
  if (straight) {
    if (flush && (sorted_cards[0] & 13) == 0) {
      /* sorted, a royal flush starts with Ace, then 10... */
      return HAND_ROYAL_FLUSH;
    } else if (flush) {
      return HAND_STRAIGHT_FLUSH;
    } else {
      return HAND_STRAIGHT;
    }
  }


  /* Find how many pairs-or-better there are */
  num_kinds = 0;
  for (i = 0; i < 5; i++) {
    cnt = 1;
    for (j = 0; j < 5; j++) {
      if (j != i) {
        if ((sorted_cards[i] % 13) == (sorted_cards[j] % 13)) {
          cnt++;
        }
      }
    }

    /* Pair or better! */
    if (cnt > 1) {
      /* See if we've already seen this value */
      found = 0;
      for (j = 0; j < num_kinds; j++) {
        if (kinds[j].val == (sorted_cards[i] % 13)) {
          found = 1;
        }
      }

      /* A new set of cards, add to the list */
      if (found == 0) {
        kinds[num_kinds].val = (sorted_cards[i] % 13);
        kinds[num_kinds].cnt = cnt;
        num_kinds++;
      }
    }
  }

  if (num_kinds == 1 && kinds[0].cnt == 4) {
    /* 4 cards? Four of a kind */
    return HAND_FOUR_KIND;
  }

  if (num_kinds == 2 && (kinds[0].cnt == 3 || kinds[1].cnt == 3)) {
    /* Two sets, and one has 3 cards, must be full house */
    return HAND_FULL_HOUSE;
  }

  if (flush) {
    /* Flush */
    return HAND_FLUSH;
  }

  if (num_kinds == 1 && kinds[0].cnt == 3) {
    /* Only one set, with 3 cards: 3 of a kind */
    return HAND_THREE_KIND;
  }

  if (num_kinds == 2) {
    /* Two sets (and not full house); 2 pairs */
    return HAND_TWO_PAIRS;
  }

  if (num_kinds == 1) {
    /* One set; 1 pair */
    return HAND_ONE_PAIR;
  }

  /* Not a scoring hand */
  return HAND_NOTHING;
}

#define GRID_TOP 10
#define GRID_LEFT 5

void print_grid() {
  unsigned char i, x, y;

  y = GRID_TOP;
  x = GRID_LEFT;
  for (i = 0; i < 25; i++) {
    print_card(x, y, grid[i]);
    if (i % 5 == 4) {
      x = GRID_LEFT;
      y += 2;
    } else {
      x += 3;
    }
  }
}

void usleep(unsigned char cnt) {
  unsigned char t1;
  unsigned int t2;

  t1 = RTCLOK_lo;
  t2 = (unsigned char) (((unsigned int) t1 + (unsigned int) cnt) % 255);
  do { } while (RTCLOK_lo != t2);
}

void play_hand_sound(unsigned char hand_score) {
  unsigned char v;

  if (hand_score != HAND_NOTHING) {
    for (v = 0; v < 15; v++) {
      SOUND(0, hand_scores[hand_score], 12, 15 - v);
      SOUND(1, hand_scores[hand_score] + 1, 12, 15 - v);
   
      usleep(hand_score);
    }
    SOUND_INIT();
  }
}

void eat_input(void) {
  do {
  } while (STICK0 != 15 || STRIG0 != 1);
  CH = KEYCODE_NONE;
}

unsigned char get_input(void) {
  unsigned char ch, s;

  eat_input();

  do {
    ch = CH & 63; /* strip Shift + Ctrl bits off */

    /* Convert any joystick input into keypresses we want */
    s = STICK0;
    if (s == 7) {
      ch = KEYCODE_ASTERISK;
    } else if (s == 11) {
      ch = KEYCODE_PLUS;
    } else if (s == 14) {
      ch = KEYCODE_MINUS;
    } else if (s == 13) {
      ch = KEYCODE_EQUALS;
    }

    if (STRIG0 == 0) {
      ch = KEYCODE_RETURN;
    }
  } while (ch != KEYCODE_RETURN &&
           ch != KEYCODE_MINUS && /* up */
           ch != KEYCODE_EQUALS && /* down */
           ch != KEYCODE_PLUS && /* left */
           ch != KEYCODE_ASTERISK /* right */);

  eat_input();

  return ch;
}

unsigned char cards[5];

int main(void) {
  unsigned int total_score, high_score;
  unsigned char cur_card, x, y, ch, i;
  unsigned char hand_score;

  SOUND_INIT();

  // _graphics(0);
  // CHBAS = ((unsigned int) &font)/256;
  SC = (unsigned char *) SAVMSC;
  LMARGN = 0;
  CRSINH = 1;

  COLOR1 = 15;
  COLOR2 = 192;

  high_score = 0;

  do {
    printf(CHS_CLR "Poker Solitaire v" VERSION " " CHS_ESC CHS_CURS_RIGHT " %u\n", high_score);
    printf("(c) Bill Kendrick 2015");
  
    shuffle_deck();
  
    cur_card = 0;
    x = 0;
    y = 0;

    for (i = 0; i < 25; i++) {
      grid[i] = 0;
    }  
    print_grid();
  
    do {
      /* Print next card */
      print_card(20, GRID_TOP + 10, deck[cur_card]);
  
      do {
        /* Draw cursor */
        ROWCRS = GRID_TOP + (y * 2) + 1;
        COLCRS = GRID_LEFT + (x * 3);
        printf(CHS_ESC CHS_CURS_UP CHS_ESC CHS_CURS_UP);
  
        /* Wait for understood keypress */
        ch = get_input();
  
        /* Erase cursor */
        ROWCRS = GRID_TOP + (y * 2) + 1;
        COLCRS = GRID_LEFT + (x * 3);
        printf("  ");
  
        /* Move cursor, if applicable */
        if (ch == KEYCODE_MINUS) {
          y = (y + 4) % 5;
        } else if (ch == KEYCODE_EQUALS) {
          y = (y + 1) % 5;
        } else if (ch == KEYCODE_PLUS) {
          x = (x + 4) % 5;
        } else if (ch == KEYCODE_ASTERISK) {
          x = (x + 1) % 5;
        }
      } while (ch != KEYCODE_RETURN);
  
      /* Place current card from deck onto the grid */
      if (grid[y * 5 + x] == 0) {
        SOUND(0, 200, 10, 8);
        grid[y * 5 + x] = deck[cur_card];
        print_card(GRID_LEFT + x * 3, GRID_TOP + y * 2, deck[cur_card]);
        cur_card++;
        SOUND(0, 0, 0, 0);
  
        cards[0] = grid[0 * 5 + x];
        cards[1] = grid[1 * 5 + x];
        cards[2] = grid[2 * 5 + x];
        cards[3] = grid[3 * 5 + x];
        cards[4] = grid[4 * 5 + x];
        hand_score = calc_hand(cards);
  
        if (hand_score != HAND_NA) {
          ROWCRS = GRID_TOP - 2;
          COLCRS = GRID_LEFT + x * 3;
          printf("%d", hand_scores[hand_score]);
  
          ROWCRS = GRID_TOP - 3 - 4 + x;
          COLCRS = GRID_LEFT + x * 3;
          printf("%s", hand_names[hand_score]); 
  
          for (i = GRID_TOP - 3 - 4 + x + 1; i <= GRID_TOP - 3; i++) {
            ROWCRS = i;
            COLCRS = GRID_LEFT + x * 3;
            printf("|");
          }

          play_hand_sound(hand_score);
        }
  
        cards[0] = grid[y * 5 + 0];
        cards[1] = grid[y * 5 + 1];
        cards[2] = grid[y * 5 + 2];
        cards[3] = grid[y * 5 + 3];
        cards[4] = grid[y * 5 + 4];
        hand_score = calc_hand(cards);
  
        if (hand_score != HAND_NA) {
          ROWCRS = GRID_TOP + y * 2;
          COLCRS = GRID_LEFT + 15;
          printf("%d", hand_scores[hand_score]);
  
          ROWCRS = GRID_TOP + y * 2;
          COLCRS = GRID_LEFT + 20;
          printf("%s", hand_names[hand_score]); 

	  play_hand_sound(hand_score);
        }
      }
    } while(cur_card < 25);
  
    print_card(20, GRID_TOP + 10, 0);
  
    ROWCRS = 22;
    COLCRS = 5;
    printf("GAME OVER  ");
  
    /* Determine final score */
    total_score = 0;
    for (i = 0; i < 5; i++) {
      cards[0] = grid[0 * 5 + i];
      cards[1] = grid[1 * 5 + i];
      cards[2] = grid[2 * 5 + i];
      cards[3] = grid[3 * 5 + i];
      cards[4] = grid[4 * 5 + i];
      hand_score = calc_hand(cards);
      total_score = total_score + ((int) hand_scores[hand_score]);
  
      cards[0] = grid[i * 5 + 0];
      cards[1] = grid[i * 5 + 1];
      cards[2] = grid[i * 5 + 2];
      cards[3] = grid[i * 5 + 3];
      cards[4] = grid[i * 5 + 4];
      hand_score = calc_hand(cards);
      total_score = total_score + ((int) hand_scores[hand_score]);
    }
  
    printf("%d ", total_score);

    /* High? */
    if (total_score > high_score) {
      printf("New high!");
      high_score = total_score;

      for (i = 10; i < 255; i++) {
        SOUND(0, i, 10, 15);
        SOUND(1, i - 10, 10, 8);
        COLOR4 = 255 - i;
      }
      SOUND_INIT();
    }
  
    sleep(1);
    printf("\n(press a key)");
    CH = KEYCODE_NONE;
    do { } while (CH == KEYCODE_NONE);
    CH = KEYCODE_NONE;
  }
  while (1); /* FIXME: How to quit? */

  printf(CHS_CLR);

  return(0);
}
